/*
 * Common routines supporting board initialisation.
 *
 * Copyright (C) 2005-2006 Ingenic Semiconductor, <jlwei@ingenic.cn>
 * Copyright (C) Xiangfu Liu <xiangfu.z@gmail.com>
 * Copyright (C) 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sdram.h"
#include "board.h"

unsigned long get_memory_size()
{
	unsigned int dmcr;
	unsigned int rows, cols, dw, banks;
	unsigned long size;

	dmcr = REG_EMC_DMCR;
	rows = SDRAM_ROW0 + ((dmcr & EMC_DMCR_RA_MASK) >> EMC_DMCR_RA_BIT);
	cols = SDRAM_COL0 + ((dmcr & EMC_DMCR_CA_MASK) >> EMC_DMCR_CA_BIT);
	dw = (dmcr & EMC_DMCR_BW) ? 2 : 4;
	banks = (dmcr & EMC_DMCR_BA) ? 4 : 2;

	size = (1 << (rows + cols)) * dw * banks;

	return size;
}

/* Timer routines. */

extern unsigned long timestamp;
extern unsigned long lastdec;

unsigned long get_timer_masked()
{
	unsigned long now = READ_TIMER;

	if (lastdec <= now) {
		/* normal mode */
		timestamp += (now - lastdec);
	} else {
		/* we have an overflow ... */
		timestamp += TIMER_FDATA + now - lastdec;
	}
	lastdec = now;

	return timestamp;
}

void reset_timer_masked()
{
	/* reset time */
	lastdec = READ_TIMER;
	timestamp = 0;
}

void reset_timer()
{
	reset_timer_masked ();
}

unsigned long get_timer(unsigned long base)
{
	return get_timer_masked () - base;
}

void set_timer(unsigned long t)
{
	timestamp = t;
}

void udelay(unsigned long usec)
{
	unsigned long tmo,tmp;

	/* normalize */
	if (usec >= 1000) {
		tmo = usec / 1000;
		tmo *= TIMER_HZ;
		tmo /= 1000;
	}
	else {
		if (usec >= 1) {
			tmo = usec * TIMER_HZ;
			tmo /= (1000*1000);
		}
		else
			tmo = 1;
	}

	/* check for rollover during this delay */
	tmp = get_timer (0);
	if ((tmp + tmo) < tmp )
		reset_timer_masked();  /* timer would roll over */
	else
		tmo += tmp;

	while (get_timer_masked () < tmo);
}

void udelay_masked(unsigned long usec)
{
	unsigned long tmo;
	unsigned long endtime;
	signed long diff;

	/* normalize */
	if (usec >= 1000) {
		tmo = usec / 1000;
		tmo *= TIMER_HZ;
		tmo /= 1000;
	} else {
		if (usec > 1) {
			tmo = usec * TIMER_HZ;
			tmo /= (1000*1000);
		} else {
			tmo = 1;
		}
	}

	endtime = get_timer_masked () + tmo;

	do {
		unsigned long now = get_timer_masked ();
		diff = endtime - now;
	} while (diff >= 0);
}

/*
 * This function is derived from PowerPC code (read timebase as long long).
 * On MIPS it just returns the timer value.
 */
unsigned long long get_ticks()
{
	return get_timer(0);
}

/*
 * This function is derived from PowerPC code (timebase clock frequency).
 * On MIPS it returns the number of timer ticks per second.
 */
unsigned long get_tbclk()
{
	return TIMER_HZ;
}

/* GPIO interrupt administration. */

void gpio_clear(uint8_t gpio)
{
	/* Clear interrupt status. */

	__gpio_ack_irq(gpio);
	__intc_ack_irq(GPIO_IRQ);
}

/* Miscellaneous interrupt administration. */

void irq_clear()
{
	uint8_t i;

	for (i = 0; i < 32; i++) {
		if (REG_INTC_IPR & (1 << i))
			__intc_ack_irq(i);
	}
}
