/*
 * Task handling.
 *
 * Copyright (C) 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cpu.h"
#include "cpu_op.h"
#include "memory.h"
#include "mips.h"
#include "paging.h"
#include "tasks.h"
#include "tasks/manifest.h"

/* Task tables and data. */

static uint32_t stack_pointers[max_tasks];
static uint32_t registers[max_tasks][32];

uint8_t current_task;
uint32_t *current_stack_pointer;
uint32_t *current_registers;

/* A reference to locations for the symbol tables. */

extern uint32_t _got_start, _got_copy_start, _got_copy_end;

/* A reference to the start of the payload and end of memory locations. */

extern uint32_t _payload_start, _memory_end;

/* Task management functions. */

void init_tasks()
{
	current_task = 1;
	init_task();
}

void init_task()
{
	current_stack_pointer = &stack_pointers[current_task];
	current_registers = registers[current_task];
}

void start_tasks()
{
	void (**starter)(uint8_t);
	int i;

	for (i = 1, starter = initial_tasks; *starter; i++, starter++)
		(*starter)(i);
}

void start_task(uint8_t task, void (*function)(), uint32_t args[], uint8_t nargs)
{
	uint32_t virtual, physical, address;

	/*
	Each task employs a stack at a multiple of the given start address in
	physical memory, but at the same address in virtual memory. Task zero
	is never started.
	*/

	virtual = TASK_STACK_TOP;

	/* NOTE: Should allocate pages generally, not according to this simple calculation. */

	physical = TASK_STACK_PHYSICAL - TASK_STACK_SIZE * task;

	init_page_table(STAGE2_PAGE_TABLE_BASE,
		previous_page(virtual, STAGE2_PAGESIZE),
		previous_page(physical, STAGE2_PAGESIZE),
		STAGE2_PAGESIZE, TLB_WRITE, task);

	/*
	Subtract from the stack pointer to prevent the called function from
	reaching into unmapped memory.
	*/

	stack_pointers[task] = virtual - 12;

	/*
	Set the registers for the new task, initialising the global pointer and
	return address.
	*/

	init_registers(registers[task], (uint32_t) &_got_copy_start, function, args, nargs);

	/* Map the global object table for the task. */

	init_page_table(STAGE2_PAGE_TABLE_BASE,
		user_address((uint32_t) &_got_start),
		user_address((uint32_t) &_got_copy_start),
		STAGE2_PAGESIZE, TLB_READ, task);

	/* Map all shared pages for the task. First, the read-only code region. */

	for (address = (uint32_t) &_payload_start;
		address < (uint32_t) &_got_start;
		address = next_page(address, STAGE2_PAGESIZE))
	{
		init_page_table(STAGE2_PAGE_TABLE_BASE,
			user_address(address),
			user_address(address),
			STAGE2_PAGESIZE, TLB_READ, task);
	}

	/* Make the pages after the code writable. */

	for (address = next_page((uint32_t) &_got_copy_end, STAGE2_PAGESIZE);
		address < (uint32_t) &_memory_end;
		address = next_page(address, STAGE2_PAGESIZE))
	{
		init_page_table(STAGE2_PAGE_TABLE_BASE,
			user_address(address),
			user_address(address),
			STAGE2_PAGESIZE, TLB_WRITE, task);
	}
}

void start_task_now()
{
	invoke_task(current_task, current_registers, current_stack_pointer);
}

void switch_task()
{
	/* Switch the current task. */

	current_task++;
	if ((current_task == max_tasks) || !(initial_tasks[current_task - 1]))
		current_task = 1;
	init_task();
}
